import os
import os.path
import json
from typing import Optional, Set

import numpy as np
from PyQt5 import QtCore
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QMainWindow, QFileDialog

from product.mode import Frequency, Voltage, Current, CurveState
from screenshot import ScreenshotFileManager
from vision.detection import dump_elements
from vision.detection import load_elements
from element import Element
from element import Pin
from oscilloscope.oscilloscopemeasurer import compare_ivc_ex
from audio.wavplayer import WavPlayer, WavFile, WavThread
from time import time


__all__ = ["MainWindow"]


class WithoutSound:

    def __init__(self, mainwin):
        self.mainwin = mainwin
        self.enable_sound_old = mainwin.enable_sound

    def __enter__(self):
        self.mainwin.set_enable_sound(False, notify_derived=False)

    def __exit__(self, *args):
        self.mainwin.enable_sound = self.enable_sound_old


class MainWindow(QMainWindow):
    """
    Base class for main window. Additionally manages trash
    with work modes and which of them are available.
    """

    score_changed = QtCore.pyqtSignal(float)
    score_threshold_changed = QtCore.pyqtSignal(float)

    enable_sound_support_changed = QtCore.pyqtSignal(bool)
    enable_sound_changed = QtCore.pyqtSignal(bool)

    frequency_changed = QtCore.pyqtSignal(Frequency)
    frequency_enabled_changed = QtCore.pyqtSignal(set)

    voltage_changed = QtCore.pyqtSignal(Voltage)
    voltage_enabled_changed = QtCore.pyqtSignal(set)

    current_changed = QtCore.pyqtSignal(Current)
    current_enabled_changed = QtCore.pyqtSignal(set)

    reference_curve_state_changed = QtCore.pyqtSignal(CurveState)
    test_curve_state_changed = QtCore.pyqtSignal(CurveState)
    external_freeze_state_changed = QtCore.pyqtSignal(CurveState)

    screenshots_allowed_changed = QtCore.pyqtSignal(bool)

    debug_changed = QtCore.pyqtSignal(bool)
    fps_changed = QtCore.pyqtSignal(float)

    auto_calibration = QtCore.pyqtSignal()

    def __init__(
            self,
            screenshot_file_manager: ScreenshotFileManager,
            parent=None,
            *args,

            score: float = 0.0,
            score_threshold: float = 0.0,
            enable_score: bool = True,

            enable_sound_support: bool = False,
            enable_sound: bool = False,

            frequency_enabled: Optional[Set[Frequency]] = None,
            frequency: Frequency = Frequency.FREQUENCY_10HZ,

            voltage_enabled: Optional[Set[Voltage]] = None,
            voltage: Voltage = Voltage.VOLTAGE_3_3V,

            current_enabled: Optional[Set[Current]] = None,
            current: Current = Current.CURRENT_250UA,

            reference_curve_state: CurveState = CurveState.NORMAL,
            test_curve_state: CurveState = CurveState.NORMAL,
            external_freeze_state: CurveState = CurveState.NORMAL,

            screenshots_allowed: bool = True,
            screenshots_show_save_as_dialog: bool = True,

            debug: bool = False,
            fps: float = 0.0
    ) -> None:
        super().__init__(parent=parent)

        if len(args) != 0:
            raise ValueError("Positional arguments not supported")

        super().__init__(parent=parent)

        self.__screenshot_file_manager = screenshot_file_manager

        self.__tab_name = "S"
        self.__wav_tabs = ["ZP", "TP"]
        self.__compare_wav = WavFile("compare.wav")
        self.__test_wav = WavFile("test.wav")
        self.__wav_last_time = 0.0
        self.__tab_curve_match = False
        self.wav_player = WavPlayer(self.__test_wav)
        self.wav_thread = WavThread(self.__compare_wav)

        self.__score = score
        self.__score_threshold = score_threshold
        self.__enable_score = enable_score

        self.__enable_sound_support = enable_sound_support
        self.__enable_sound = enable_sound

        self.__frequency = frequency
        self.__frequency_prev = frequency
        self.__frequency_enabled = frequency_enabled\
            if frequency_enabled is not None else set(f for f in Frequency)
        self.__frequency_enabled_original = set(self.__frequency_enabled)

        self.__voltage = voltage
        self.__voltage_prev = voltage
        self.__voltage_enabled = voltage_enabled\
            if voltage_enabled is not None else set(v for v in Voltage)
        self.__voltage_enabled_original = set(self.__voltage_enabled)

        self.__current = current
        self.__current_prev = current
        self.__current_enabled = current_enabled\
            if current_enabled is not None else set(c for c in Current)
        self.__current_enabled_original = set(self.__current_enabled)

        self.__reference_curve_state = reference_curve_state
        self.__test_curve_state = test_curve_state
        self.__external_freeze_state = external_freeze_state
        self.__reference_curve_state_backup = self.__reference_curve_state
        self.__test_curve_state_backup = self.__test_curve_state

        self.__screenshots_allowed = screenshots_allowed
        self.__screenshots_show_save_as_dialog =\
            screenshots_show_save_as_dialog

        self.__debug = debug
        self.__fps = fps
        self.__reference_curve = None
        self.__test_curve = None
        self.__reference_curve_override = None
        self.__test_curve_override = None
        self.__devices_connected = False
        self.__file_not_found = False
        self.__test_plan_pin_index = 0

    def __refresh_main_widget(self):
        visible = self.ivc_viewer
        if not self.__devices_connected:
            visible = self.label_DeviceStatus
        elif self.__file_not_found:
            visible = self.label_FileNotFound
        for wdg in (
                self.ivc_viewer,
                self.label_DeviceStatus,
                self.label_FileNotFound):
            wdg.setVisible(wdg is visible)

    def set_devices_ready(self, connected):
        self.__devices_connected = connected
        self.__refresh_main_widget()

    def __play_match_wav(self):
        match = self.__score < self.__score_threshold
        if self.enable_sound_generator:
            self.wav_thread.play = (not match) and self.enable_sound
        elif match != self.__tab_curve_match:
            self.wav_thread.play = False
            self.__tab_curve_match = match
            if match and self.enable_sound and not self.enable_sound_generator:
                now = time()
                count = 1
                diff = now - self.__wav_last_time
                if diff >= self.__test_wav.duration * count:
                    self.__wav_last_time = now
                    self.wav_player.play(count=count)

    def _set_curves(
            self,
            reference_curve, test_curve,
            frequency: Frequency, voltage: Voltage, current: Current
    ) -> None:
        pass

    def set_curves(
            self,
            reference_curve, test_curve,
            frequency: Frequency, voltage: Voltage, current: Current
    ) -> None:
        self.__reference_curve = reference_curve
        self.__test_curve = test_curve
        if self.__reference_curve_override is not None:
            reference_curve = self.__reference_curve_override
        if self.__test_curve_override is not None:
            test_curve = self.__test_curve_override
        self._set_curves(
            reference_curve,
            test_curve,
            frequency,
            voltage,
            current
        )
        if self.__enable_score:
            self.__score = compare_ivc_ex(
                reference_curve,
                test_curve,
                self.__voltage,
                self.__current
            )
            self.__play_match_wav()
            self._set_score(self.__score)
            self.score_changed.emit(self.__score)

    def _set_score(self, score: float) -> None:
        pass

    @property
    def score(self):
        return self.__score

    def _set_score_threshold(self, score_threshold: float) -> None:
        pass

    def __refresh_score_threshold_gui(self):
        value = self.score_threshold
        if value < 0.0:
            value = 0.0
        if value > 1.0:
            value = 1.0
        self.pushButton_score_threshold_minus.setEnabled(value > 0.0)
        self.pushButton_score_threshold_plus.setEnabled(value < 1.0)
        self.label_score_threshold_value.setText(
            "{value:.2f}".format(value=self.score_threshold)
        )

    def on_score_threshold_button(self, add):
        if add:
            self.score_threshold = min(1.0, self.score_threshold + 0.05)
        else:
            self.score_threshold = max(0.0, self.score_threshold - 0.05)
        self.__refresh_score_threshold_gui()

    def get_score_threshold(self) -> float:
        return self.__score_threshold

    @QtCore.pyqtSlot(float)
    def set_score_threshold(self, score_threshold: float) -> None:
        self.__score_threshold = score_threshold
        self._set_score_threshold(self.__score_threshold)
        self.score_threshold_changed.emit(self.__score_threshold)

    score_threshold = property(get_score_threshold, set_score_threshold)

    def _set_enable_sound_support(self, enable_sound_support: bool) -> None:
        pass

    def get_enable_sound_support(self) -> bool:
        return self.__enable_sound_support

    @QtCore.pyqtSlot(bool)
    def set_enable_sound_support(self, enable_sound_support: bool) -> None:
        self.__enable_sound_support = enable_sound_support
        self._set_enable_sound_support(self.__enable_sound_support)
        self.enable_sound_support_changed.emit(self.__enable_sound_support)

    enable_sound_support = property(
        get_enable_sound_support, set_enable_sound_support)

    def _set_enable_sound(self, enable_sound: bool) -> None:
        pass

    def get_enable_sound(self) -> bool:
        return self.__enable_sound

    @QtCore.pyqtSlot(bool)
    def set_enable_sound(self, enable_sound: bool, notify_derived=True) -> None:
        self.__enable_sound = enable_sound
        if notify_derived:
            self._set_enable_sound(self.__enable_sound)
        self.enable_sound_changed.emit(self.__enable_sound)

    enable_sound = property(get_enable_sound, set_enable_sound)

    @property
    def enable_sound_generator(self) -> bool:
        return self.__tab_name not in self.__wav_tabs

    def _set_frequency(self, frequency: Frequency) -> None:
        pass

    def get_frequency(self) -> Frequency:
        return self.__frequency

    @QtCore.pyqtSlot(Frequency)
    def set_frequency(self, frequency: Frequency) -> None:
        self.__frequency = frequency
        self._set_frequency(self.__frequency)
        self.frequency_changed.emit(self.__frequency)

    frequency = property(get_frequency, set_frequency)

    def _set_frequency_enabled(
            self, frequency_enabled: Optional[Set[Frequency]]) -> None:
        pass

    def get_frequency_enabled(self) -> Set[Frequency]:
        return self.__frequency_enabled

    @QtCore.pyqtSlot(Set[Frequency])
    def set_frequency_enabled(
            self, frequency_enabled: Optional[Set[Frequency]]) -> None:
        self.__frequency_enabled = frequency_enabled\
            if frequency_enabled is not None else set(f for f in Frequency)
        self._set_frequency_enabled(self.__frequency_enabled)
        self.frequency_enabled_changed.emit(self.__frequency_enabled)

    @QtCore.pyqtSlot()
    def clear_frequency_enabled(self):
        self.__frequency_enabled = set(f for f in Frequency)

    frequency_enabled = property(
        get_frequency_enabled,
        set_frequency_enabled,
        clear_frequency_enabled
    )

    def _set_voltage(self, voltage: Voltage) -> None:
        pass

    def get_voltage(self) -> Voltage:
        return self.__voltage

    @QtCore.pyqtSlot(Voltage)
    def set_voltage(self, voltage: Voltage) -> None:
        self.__voltage = voltage
        self._set_voltage(self.__voltage)
        self.voltage_changed.emit(self.__voltage)

    voltage = property(get_voltage, set_voltage)

    def _set_voltage_enabled(
            self, voltage_enabled: Optional[Set[Voltage]]) -> None:
        pass

    def get_voltage_enabled(self) -> Set[Voltage]:
        return self.__voltage_enabled

    @QtCore.pyqtSlot(Set[Voltage])
    def set_voltage_enabled(
            self, voltage_enabled: Optional[Set[Voltage]]) -> None:
        self.__voltage_enabled = voltage_enabled\
            if voltage_enabled is not None else set(f for f in Voltage)
        self._set_voltage_enabled(self.__voltage_enabled)
        self.voltage_enabled_changed.emit(self.__voltage_enabled)

    @QtCore.pyqtSlot()
    def clear_voltage_enabled(self):
        self.__voltage_enabled = set(f for f in Voltage)

    voltage_enabled = property(
        get_voltage_enabled, set_voltage_enabled, clear_voltage_enabled)

    def _set_current(self, current: Current) -> None:
        pass

    def get_current(self) -> Current:
        return self.__current

    @QtCore.pyqtSlot(Current)
    def set_current(self, current: Current) -> None:
        self.__current = current
        self._set_current(self.__current)
        self.current_changed.emit(self.__current)

    current = property(get_current, set_current)

    def _set_current_enabled(
            self, current_enabled: Optional[Set[Current]]) -> None:
        pass

    def get_current_enabled(self) -> Set[Current]:
        return self.__current_enabled

    @QtCore.pyqtSlot(Set[Current])
    def set_current_enabled(
            self, current_enabled: Optional[Set[Current]]) -> None:
        self.__current_enabled = current_enabled\
            if current_enabled is not None else set(f for f in Current)
        self._set_current_enabled(self.__current_enabled)
        self.current_enabled_changed.emit(self.__current_enabled)

    @QtCore.pyqtSlot()
    def clear_current_enabled(self):
        self.__current_enabled = set(f for f in Current)

    current_enabled = property(
        get_current_enabled, set_current_enabled, clear_current_enabled)

    def _set_reference_curve_state(
            self, reference_curve_state: CurveState) -> None:
        pass

    def get_reference_curve_state(self) -> CurveState:
        return self.__reference_curve_state

    @QtCore.pyqtSlot(CurveState)
    def set_reference_curve_state(
            self, reference_curve_state: CurveState) -> None:
        if self.__external_freeze_state == CurveState.NORMAL:
            self.__reference_curve_state = reference_curve_state
            self._set_reference_curve_state(self.__reference_curve_state)
            self.reference_curve_state_changed.emit(
                self.__reference_curve_state)

    reference_curve_state = property(
        get_reference_curve_state, set_reference_curve_state)

    def _set_test_curve_state(self, test_curve_state: CurveState) -> None:
        pass

    def get_test_curve_state(self) -> CurveState:
        return self.__test_curve_state

    @QtCore.pyqtSlot(CurveState)
    def set_test_curve_state(self, test_curve_state: CurveState) -> None:
        if self.__external_freeze_state == CurveState.NORMAL:
            self.__test_curve_state = test_curve_state
            self._set_test_curve_state(self.__test_curve_state)
            self.test_curve_state_changed.emit(self.__test_curve_state)

    test_curve_state = property(get_test_curve_state, set_test_curve_state)

    def _set_external_freeze_state(
            self, external_freeze_state: CurveState) -> None:
        pass

    def get_external_freeze_state(self) -> CurveState:
        return self.__external_freeze_state

    @QtCore.pyqtSlot(CurveState)
    def set_external_freeze_state(
            self, external_freeze_state: CurveState) -> None:
        if self.__external_freeze_state != external_freeze_state:
            self.__external_freeze_state = external_freeze_state
            if self.__external_freeze_state == CurveState.NORMAL:
                self.__reference_curve_state =\
                    self.__reference_curve_state_backup
                self.__test_curve_state = self.__test_curve_state_backup
            else:
                self.__reference_curve_state_backup =\
                    self.__reference_curve_state
                self.__reference_curve_state = CurveState.FROZEN
                self.__test_curve_state_backup = self.__test_curve_state
                self.__test_curve_state = CurveState.FROZEN

            self._set_external_freeze_state(self.__external_freeze_state)
            self.external_freeze_state_changed.emit(
                self.__external_freeze_state)

    external_freeze_state = property(
        get_external_freeze_state, set_external_freeze_state)

    def _set_screenshots_allowed(self, screenshots_allowed: bool) -> None:
        pass

    def get_screenshots_allowed(self) -> bool:
        return self.__screenshots_allowed

    @QtCore.pyqtSlot(bool)
    def set_screenshots_allowed(self, screenshots_allowed: bool) -> None:
        self.__screenshots_allowed = screenshots_allowed
        self._set_screenshots_allowed(screenshots_allowed)
        self.screenshots_allowed_changed.emit(self.__screenshots_allowed)

    screenshots_allowed = property(
        get_screenshots_allowed, set_screenshots_allowed)

    def _set_debug(self, debug: bool) -> None:
        pass

    def get_debug(self) -> bool:
        return self.__debug

    @QtCore.pyqtSlot(bool)
    def set_debug(self, debug: bool) -> None:
        self.__debug = debug
        self._set_debug(self.__debug)
        self.debug_changed.emit(self.__debug)

    debug = property(get_debug, set_debug)

    def _set_fps(self, fps: float) -> None:
        pass

    def get_fps(self) -> float:
        return self.__fps

    @QtCore.pyqtSlot(float)
    def set_fps(self, fps: float) -> None:
        self.__fps = fps
        self._set_fps(self.__fps)
        self.fps_changed.emit(self.__fps)

    fps = property(get_fps, set_fps)

    def _make_screenshot(self) -> Optional[QPixmap]:
        pass

    @QtCore.pyqtSlot()
    def make_screenshot(self) -> None:
        with WithoutSound(self):
            screenshot = self._make_screenshot()
            if screenshot is None:
                return

            screenshot_path = self.__screenshot_file_manager\
                .get_next_screenshot_file_name()
            if screenshot_path is None:
                return

            if self.__screenshots_show_save_as_dialog:
                screenshot_path = QFileDialog.getSaveFileName(
                    parent=self,
                    caption="Save screen shot",
                    directory=screenshot_path,
                    filter="Images (*.png)"
                )[0]

                if len(screenshot_path) == 0:
                    return

            screenshot.save(screenshot_path)
            self.__screenshot_file_manager.commit_screenshot()
            print("Screenshot saved to: \"{}\".".format(screenshot_path))

    def on_freeze_gpio(self, high):
        with WithoutSound(self):
            tab_index = self.test_plan_tab_widget.currentIndex()
            if tab_index == 0:
                self.set_external_freeze_state(
                    CurveState.FROZEN if high else CurveState.NORMAL)
            elif tab_index == 1:
                if high:
                    self.on_test_plan_zp_save()
            elif tab_index == 2:
                if high:
                    self.__test_plan_pin_refresh(1)

    def __get_test_plan_element(self):
        with WithoutSound(self):
            self.__file_not_found = True
            reference_path = self.__screenshot_file_manager.get_reference_path()
            if reference_path is None:
                return None
            try:
                if not os.path.isdir(reference_path):
                    os.makedirs(reference_path, exist_ok=True)
            except Exception as ex:
                print("Failed to make dirs: {}.".format(reference_path))
                print(ex)
                return None

            board_path = os.path.join(reference_path, "board.json")
            error = None
            try:
                with open(board_path, "r") as f:
                    json.load(f)
            except FileNotFoundError:
                error = "not found"
            except ValueError:
                error = "bad json"
            except Exception:
                error = "unknown exception"
            if error is not None:
                print("{path} {error} - creating...".format(
                    path=board_path,
                    error=error
                ))
                element = None
            else:
                try:
                    element = load_elements(board_path)[0]
                    self.__file_not_found = False
                except Exception:
                    element = None
            return board_path, element

    def __test_plan_pin_refresh(self, index_diff=0):
        with WithoutSound(self):
            path_elem = self.__get_test_plan_element()
            if path_elem is None:
                self.__file_not_found = True
                return
            board_path, element = path_elem
            if element is None or len(element.pins) == 0:
                self.__file_not_found = True
                return
            self.__file_not_found = False
            pins = element.pins
            self.__test_plan_pin_index += index_diff
            if self.__test_plan_pin_index < 0:
                self.__test_plan_pin_index = 0
            if self.__test_plan_pin_index >= len(element.pins):
                self.__test_plan_pin_index = len(element.pins) - 1
            pin = element.pins[self.__test_plan_pin_index]
    #       self.__reference_curve_override = pin.reference_ivc
            self.__test_curve_override = pin.ivc
            self.__refresh_main_widget()
            self.test_plan_ui_refresh(
                index=self.__test_plan_pin_index,
                left=(element is not None) and (self.__test_plan_pin_index > 0),
                right=(element is not None) and (self.__test_plan_pin_index < (len(pins) - 1)),
                file_found=not self.__file_not_found
            )
            self.__frequency_prev = self.__frequency
            self.__voltage_prev = self.__voltage
            self.__current_prev = self.__current
            self.frequency = pin.frequency
            self.voltage = pin.voltage
            self.current = pin.current

    def __test_plan_restore(self):
        self.frequency = self.__frequency_prev
        self.voltage = self.__voltage_prev
        self.current = self.__current_prev

    def __tab_sound_setup(self, name):
        self.__tab_name = name
        self.__tab_curve_match = False

    def on_test_plan_tab_s(self):
        self.__tab_sound_setup("S")
        self.__test_plan_restore()
        self.set_frequency_enabled(self.__frequency_enabled_original)
        self.set_voltage_enabled(self.__voltage_enabled_original)
        self.set_current_enabled(self.__current_enabled_original)
        self.__file_not_found = False
        self.__reference_curve_override = None
        self.__test_curve_override = None

    def __test_plan_create_pin(self, only_if_empty):
        with WithoutSound(self):
            path_elem = self.__get_test_plan_element()
            if path_elem is None:
                self.__file_not_found = True
                return
            board_path, element = path_elem
            do_save = False
            if element is None:
                element = Element(
                    pins=[],
                    is_manual=True,
                    center=np.zeros(2)
                )
                do_save = True
            if not only_if_empty or len(element.pins) == 0:
                pin = self.__make_pin(default_values=True)
                if pin is None:
                    return
                element.pins.append(pin)
                self.__test_plan_pin_index = len(element.pins) - 1
                self.frequency = pin.frequency
                self.voltage = pin.voltage
                self.current = pin.current
                do_save = True
            if do_save:
                dump_elements(board_path, [element])
            self.__test_plan_pin_refresh()

    def __make_pin(self, default_values):
        with WithoutSound(self):
            if self.__reference_curve is None:
                return None
            # if self.__test_curve is None:
            #    return None
            if default_values:
                ms = self.__reference_curve[0]
                ivc = (
                    ms,
                    (
                        # Sin probe signal
                        np.sin(np.linspace(0, 2 * np.pi, ms.n_points)),
                        np.zeros(ms.n_points)
                    )
                )
            else:
                ivc = self.__reference_curve
            return Pin(
                rc=np.zeros(2),
                ivc=ivc,
                reference_ivc=None,
                frequency=self.frequency,
                voltage=self.voltage,
                current=self.current
            )

    def on_test_plan_tab_zp(self):
        self.__tab_sound_setup("ZP")
        self.reference_curve_state = CurveState.NORMAL
        self.test_curve_state = CurveState.NORMAL
        self.set_frequency_enabled(self.__frequency_enabled_original)
        self.set_voltage_enabled(self.__voltage_enabled_original)
        self.set_current_enabled(self.__current_enabled_original)
        self.__test_plan_restore()
        self.__test_plan_create_pin(only_if_empty=True)

    def on_test_plan_zp_left(self):
        self.__test_plan_pin_refresh(-1)

    def on_test_plan_zp_right(self):
        self.__test_plan_pin_refresh(1)

    def on_test_plan_zp_save(self):
        with WithoutSound(self):
            pin = self.__make_pin(default_values=False)
            if pin is None:
                return
            path_elem = self.__get_test_plan_element()
            if path_elem is None:
                self.__file_not_found = True
                return
            board_path, element = path_elem
            if element is not None:
                element.pins[self.__test_plan_pin_index] = pin
                dump_elements(board_path, [element])
            self.__test_plan_pin_refresh(0)

    def on_test_plan_zp_new_point(self):
        self.__test_plan_create_pin(only_if_empty=False)

    def on_test_plan_tab_tp(self):
        with WithoutSound(self):
            self.__tab_sound_setup("TP")
            self.reference_curve_state = CurveState.NORMAL
            self.test_curve_state = CurveState.NORMAL
            self.__test_plan_pin_refresh()
            self.set_frequency_enabled(set())
            self.set_voltage_enabled(set())
            self.set_current_enabled(set())

    def on_test_plan_tp_left(self):
        self.__test_plan_pin_refresh(-1)

    def on_test_plan_tp_right(self):
        self.__test_plan_pin_refresh(1)

    def on_test_plan_tp_save(self):
        self.make_screenshot()

    def on_test_plan_tab_c(self):
        with WithoutSound(self):
            self.__tab_sound_setup("C")
            self.__test_plan_restore()
            self.set_frequency_enabled(self.__frequency_enabled_original)
            self.set_voltage_enabled(self.__voltage_enabled_original)
            self.set_current_enabled(self.__current_enabled_original)
            self.__file_not_found = False
            self.__reference_curve_override = None
            self.__test_curve_override = None
            self.__refresh_score_threshold_gui()

    def on_test_plan_auto_calibration(self):
        with WithoutSound(self):
            self.auto_calibration.emit()
