﻿import numpy as np
from collections import namedtuple
from scipy import interpolate


__all__ = [
    "MeasureSettings",
    "dist2_pt_seg",
    "rescale_score",
    "dist_curve_pts_slow",
    "dist_curve_pts_rude",
    "dist_curve_pts",
    "remove_repeats_ivc",
    "compare_ivc",
    "Oscilloscope"
]


class MeasureSettings(
    namedtuple(
        "MeasureSettings",
        (
            "desc_frequency",
            "probe_signal_frequency",
            "n_points",
            "n_charge_points",
            "max_voltage",
            "max_current",
            "flags",
            "reserved"
        )
    )
):
    def to_json(self) -> dict:
        return {
            "desc_frequency": float(self.desc_frequency),
            "probe_signal_frequency": float(self.probe_signal_frequency),
            "n_points": int(self.n_points),
            "n_charge_points": int(self.n_charge_points),
            "max_voltage": float(self.max_voltage),
            "max_current": float(self.max_current),
            "flags": int(self.flags),
            "reserved": tuple(map(int, self.reserved))
        }

    @classmethod
    def from_json(cls, d: dict) -> "MeasureSettings":
        return cls(
            desc_frequency=d["desc_frequency"],
            probe_signal_frequency=d["probe_signal_frequency"],
            n_points=d["n_points"],
            n_charge_points=d["n_charge_points"],
            max_voltage=d["max_voltage"],
            max_current=d["max_current"],
            flags=d["flags"],
            reserved=np.array(d["reserved"], dtype=np.uint8)
        )


eq_k = 1


def dist2_pt_seg(p, a, b):
    v1 = np.subtract(b, a)
    v2 = np.subtract(p, a)
    seg_len2 = np.dot(v1, v1)
    proj = np.dot(v1, v2) / seg_len2
    if proj < 0:
        return np.dot(v2, v2)
    if proj > 1:
        return np.dot(np.subtract(p, b), np.subtract(p, b))

    return np.cross(v1, v2) ** 2 / seg_len2


def dist2_pt_seg_vec(p, a, b):
    # Vectorized dist2_pt_seg
    # Send vectors like [[x1, y1], [x2, y2], ...]
    # instead of points
    p = np.array(p)
    a = np.array(a)
    b = np.array(b)

    v1_x = b[:, 0] - a[:, 0]
    v1_y = b[:, 1] - a[:, 1]

    v2_x = p[:, 0] - a[:, 0]
    v2_y = p[:, 1] - a[:, 1]

    v3_x = p[:, 0] - b[:, 0]
    v3_y = p[:, 1] - b[:, 1]

    # seg_len2 = np.dot(v1, v1)
    seg_len2 = v1_x * v1_x + v1_y * v1_y

    # proj = np.dot(v1, v2) / seg_len2
    proj = np.where(
        seg_len2 != 0, (v1_x * v2_x + v1_y * v2_y) / seg_len2,
        np.inf
    )

    # z projection of [v1 x v2]
    vc = v1_x * v2_y - v1_y * v2_x
    vc_n = np.where(seg_len2 != 0, vc ** 2 / seg_len2, np.inf)

    # Magic legacy
    res_1 = np.where(proj < 0, v2_x * v2_x + v2_y * v2_y, 0)
    res_2 = np.where(proj > 1, v3_x * v3_x + v3_y * v3_y, 0)
    res_3 = np.where(abs(proj - 0.5) <= 0.5, vc_n, 0)

    return res_1 + res_2 + res_3


def rescale_score(x):
    return 1 - np.exp(-8 * x)


def dist2():
    raise NotImplementedError("dist2")


def dist_curve_pts_slow(curve, pts):
    res = 0.0
    for pt in pts.T:
        min_v = np.inf
        for i in range(1, len(curve[0])):
            v = dist2_pt_seg(pt, curve[:, i - 1], curve[:, i])
            if v < min_v:
                min_v = v
                min_i = i
        res += min_v
        v = dist2(pt, curve[:, min_i])
        # if abs((v / min_v) - 1) > 0.1:
        #    print("seg, pts =", v, min_v)
    return res


def dist_curve_pts_rude(curve, pts):
    res = 0.0
    for pt in pts.T:
        minv = np.inf
        for i in range(0, len(curve[0])):
            v = dist2(pt, curve[:, i])
            # vdiff = pt[0] - curve[0, i]
            # cdiff = pt[1] - curve[1, i]
            # v = vdiff * vdiff + cdiff * cdiff
            if v < minv:
                minv = v
                # min_i = i
        res += minv
    return res


def dist_curve_pts(curve, pts):
    # curve:                [[x1, x2, ...], y[y1, y2, ...]]
    # pts - set of points:  [[x1, x2, ...], y[y1, y2, ...]]
    # in fact pts - is an other curve

    # from matplotlib import pyplot as plt
    # plt.plot(*curve)
    # plt.scatter(*pts, c="red")

    p = []
    c1 = []
    c2 = []
    c3 = []
    mask_1 = []
    mask_3 = []

    for pt in pts.T:

        # Calculate distances for all points
        # d0 = a[0] - b[0]
        # d1 = a[1] - b[1]
        # d0 * d0 + d1 * d1
        v = (curve[0] - pt[0]) * (curve[0] - pt[0]) + (
            curve[1] - pt[1]) * (curve[1] - pt[1])
        min_i = v.argmin()

        p.append(pt)

        if min_i > 0:
            c1.append(curve[:, min_i - 1])
            mask_1.append(True)
        else:
            c1.append(np.zeros_like(curve[:, 0]))
            mask_1.append(False)

        c2.append(curve[:, min_i])

        if min_i < len(curve[0]) - 1:
            c3.append(curve[:, min_i + 1])
            mask_3.append(True)
        else:
            c3.append(np.zeros_like(curve[:, 0]))
            mask_3.append(False)

    r1 = np.where(mask_1, dist2_pt_seg_vec(pts.T, c1, c2), np.inf)
    r3 = np.where(mask_3, dist2_pt_seg_vec(pts.T, c2, c3), np.inf)
    r = np.minimum(r1, r3)

    res = sum(r)

    res /= len(pts.T)
    # plots = np.reshape((pt, curve[:, ordered[min_i]]), (2, 2)).T
    # plt.plot(*plots, c="green")
    # print("res %.4f" % res)
    # plt.show()
    return res


def remove_repeats_ivc(a, eps=1e-6):
    msk0 = np.append(np.abs(a[0][1:] - a[0][:-1]) > eps, True)
    msk1 = np.append(np.abs(a[1][1:] - a[1][:-1]) > eps, True)
    return a[0][msk0 | msk1], a[1][msk0 | msk1]


def compare_ivc(a, b=None, min_var_v=None, min_var_c=None):
    # a, b:  tuple(oscilloscope instance, curve)
    # curve: tuple(voltage_points, current_points)

    if not a or a[1] is None:
        return 0

    a = a[1]
    if b is not None:
        b = b[1]

    # Now a and b - curves - tuple(voltage_points, current_points)

    if min_var_v is None:
        min_var_v = 0.6

    if min_var_c is None:
        min_var_c = 0.0002

    # The variance is the average of the squared deviations from the mean,
    # i.e., var = mean(abs(x - x.mean())**2).
    var_v = max(
        np.var(a[0]) ** 0.5,
        np.var(b[0]) ** 0.5 if b is not None else 0, min_var_v)
    var_c = max(
        np.var(a[1]) ** 0.5,
        np.var(b[1]) ** 0.5 if b is not None else 0, min_var_c)

    # Rescale curves for standard size and center them
    an = np.subtract(a[0], np.mean(a[0])) / var_v, np.subtract(
        a[1], np.mean(a[1])) / var_c

    an = remove_repeats_ivc(an)

    tck, u = interpolate.splprep(an, s=0.00)
    eq1 = np.array(interpolate.splev(
        np.arange(0, 1, 1.0 / len(a[0]) / eq_k), tck))

    if b is None:
        return rescale_score(np.mean(eq1[1, :] ** 2))
    else:
        bn = np.subtract(b[0], np.mean(b[0])) / var_v, np.subtract(
            b[1], np.mean(b[1])) / var_c
        bn = remove_repeats_ivc(bn)
        tck, u = interpolate.splprep(bn, s=0.00)
        eq2 = np.array(interpolate.splev(
            np.arange(0, 1, 1.0 / len(b[0]) / eq_k), tck))

        return rescale_score(
            (dist_curve_pts(eq1, eq2) + dist_curve_pts(eq2, eq1)) / 2.
        )


class Oscilloscope:
    def __init__(self):
        self.s = MeasureSettings(
            desc_frequency=None,
            probe_signal_frequency=None,
            max_voltage=5.,
            max_current=1.,
            n_points=100,
            n_charge_points=0,
            flags=0x13,
            reserved=np.zeros(20, dtype=np.uint8)
        )

    def open_device(self):
        pass

    def close_device(self):
        pass

    @property
    def is_opened(self):
        return True

    def measure(self):
        ang_dev = np.pi / 4
        if not hasattr(self, "curve") or self.__curve is None:
            self.__curve = np.zeros((2, self.s.n_points), dtype=np.float32)
            self.__ang = np.random.uniform(0, np.pi * 2)
            begin = 1
        else:
            self.__curve = np.roll(self.__curve, -1, axis=1)
            begin = len(self.__curve[0]) - 1

        for i in range(begin, len(self.__curve[0])):
            self.__ang += np.random.uniform(-ang_dev, ang_dev)
            length_v = self.s.max_voltage / 5
            length_c = self.s.max_current / 5
            while True:
                self.__curve[:, i] = self.__curve[:, i - 1] + (
                    length_v * np.cos(self.__ang),
                    length_c * np.sin(self.__ang)
                )
                if -self.s.max_voltage\
                        <= self.__curve[0, i]\
                        <= self.s.max_voltage and \
                        -self.s.max_current\
                        <= self.__curve[1, i] <= self.s.max_current:
                    break
                self.__ang += ang_dev
        res = self.__curve.copy()
        res[1] /= 1000.0
        return self.s, res

    def fast_calibrate(self):
        pass

    def full_calibrate(self):
        pass

    def prepareForMeasure(self):
        pass

    def set_settings(self, **kwargs):
        for f in MeasureSettings._fields:
            if f not in kwargs:
                kwargs[f] = self.s._asdict()[f]
        self.s = MeasureSettings(**kwargs)
        self.__curve = None

    def get_settings(self):
        return self.s
